from typing import *
from base_aux.funcs import ValueNotExist


# =====================================================================================================================
class GetattrAux:
    """
    NOTICE
    ------
    if there are several same attrs in different cases - you should resolve it by yourself!
    """
    @classmethod
    def _attr_anycase__get_name(cls, item: str | Any, obj: Any = ValueNotExist) -> str | None:
        """
        get attr name in original register
        """
        if obj == ValueNotExist:
            obj = cls   # seems it is not good idea!!!

        for name in dir(obj):
            if name.lower() == str(item).lower():
                return name

    @classmethod
    def _attr_anycase__get_value(cls, item: str, obj: Any) -> Any | Callable | NoReturn:
        """
        get attr value by name in any register
        no execution! return pure value as represented in object!
        """
        name_original = cls._attr_anycase__get_name(item, obj)
        if name_original is None:
            raise AttributeError(item)

        return getattr(obj, name_original)


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class ValueNotExist:
    """
    DEPRECATE???
    ---------
    use direct ArgsEmpty???

    GOAL
    ----
    it is different from Default!
    there is no value!
    used when we need to change logic with not passed value!

    SPECIALLY CREATED FOR
    ---------------------
    Valid as universal validation object under cmp other objects!

    USAGE
    -----
    class Cls:
        def __init__(self, value: Any | Type[ValueNotExist] | ValueNotExist = ValueNotExist):
            self.value = value

        def __eq__(self, other):
            if self.value is ValueNotExist:
                return other is True
                # or
                return self.__class__(other).run()
            else:
                return other == self.value

        def run(self):
            return bool(self.value)

    SAME AS
    -------
    args.ArgsEmpty but single and really not defined
    """
    pass


# =====================================================================================================================
TYPE__VALUE_NOT_PASSED = Type[ValueNotExist] | ValueNotExist

TYPE__ARGS = Union[tuple, Any, None, "TYPE__ARGS_EMPTY", "TYPE__EXPLICIT"]
TYPE__KWARGS = Optional[dict[str, Any]]

TYPE__EXCEPTION = Union[Exception, Type[Exception]]
TYPE__SOURCE_LINK = Union[Any, TYPE__EXCEPTION, Callable[[...], Any | NoReturn], TYPE__VALUE_NOT_PASSED]


# =====================================================================================================================



# =====================================================================================================================
# VERSION = (0, 0, 1)   # use import EXACT_OBJECTS! not *
#   from .main import *                 # INcorrect
#   from .main import EXACT_OBJECTS     # CORRECT
# VERSION = (0, 0, 2)   # del blank lines
# VERSION = (0, 0, 3)   # separate all types/exx into static.py!


# =====================================================================================================================
# TEMPLATE
# from .STATIC import (
#     # TYPES
#     # EXX
# )
# from .main import (
#     # BASE
#     # AUX
# )
# ---------------------------------------------------------------------------------------------------------------------
from .static import (
    # TYPES
    # EXX
    Exx__AnnotNotDefined,
)
from .annot_1_aux import (
    # BASE
    AnnotAux,
    # AUX
)
from .annot_3_iter_values import AnnotValuesIter
from .annot_2_all_defined import (
    # BASE
    AnnotAllDefined,
    # AUX
)
from .annot_4_cls_keys_as_values import (
    # BASE
    AnnotClsKeysAsValues,
    # AUX
    AnnotClsKeysAsValues_Meta,
)
# ---------------------------------------------------------------------------------------------------------------------
from .cmp import (
    # BASE
    CmpInst,
    # AUX
    # TYPES
    # EXX
)
from .number import (
    # BASE
    NumberArithmTranslateToAttr,
    # AUX
    # TYPES
    TYPE__NUMBER,
    # EXX
    Exx__NumberArithm_NoName,
)
# ---------------------------------------------------------------------------------------------------------------------
from .getattr_0_echo import (
    # BASE
    GetattrEcho,
    GetattrEchoSpace,
    # AUX
    # TYPES
    # EXX
)
from .getattr_1_aux import (
    # BASE
    GetattrAux,
    # AUX
    # TYPES
    # EXX
)
from .getattr_2_anycase import (
    # BASE
    GetattrAnycase,
    # AUX
    # TYPES
    # EXX
)
from .getattr_3_prefix_1_inst import (
    # BASE
    GetattrPrefixInst,
    GetattrPrefixInst_RaiseIf,
    # AUX
    # TYPES
    # EXX
    Exx__GetattrPrefix,
    Exx__GetattrPrefix_RaiseIf,
)
from .getattr_3_prefix_2_cls import GetattrPrefixCls_MetaTemplate

# ---------------------------------------------------------------------------------------------------------------------
from .middle_group import (
    # BASE
    ClsMiddleGroup,
    # AUX
    # TYPES
    # EXX
)

# ---------------------------------------------------------------------------------------------------------------------
from .value_1_variants import (
    # BASE
    ValueVariants,
    # AUX
    # TYPES
    # EXX
    Exx__ValueNotInVariants,
    Exx__VariantsIncompatible,
)
from .value_2_unit import (
    # BASE
    ValueUnit,
    # AUX
    UnitBase,
    UNIT_MULT__VARIANTS,
    # TYPES
    Exx__ValueNotParsed,
    Exx__ValueUnitsIncompatible,
    # EXX
)
# ---------------------------------------------------------------------------------------------------------------------
from .valid_0_aux import ValidAux
from .valid_1_base import (
    # BASE
    Valid,
    # AUX
    # TYPES
    # TYPE__EXCEPTION,
    # TYPE__SOURCE_LINK,
    TYPE__VALIDATE_LINK,
    TYPE__BOOL_LINK,
    # EXX
)
from .valid_2_chains import (
    ValidChains,
    TYPE__CHAINS,
)
from .valid_1_base_derivatives import (
    # BASE
    ValidRetry1,
    ValidRetry2,
    ValidFailStop,
    ValidFailContinue,
    ValidNoCum,
    ValidReverse,
    ValidSleep,
    # AUX
    # TYPES
    # EXX
)
from .valid_3_regexp import (
    # BASE
    ValidRegExp,
    # AUX
    # TYPES
    # EXX
)

# ---------------------------------------------------------------------------------------------------------------------
from .breeder_1_str_1_series import (
    # BASE
    BreederStrSeries,
    # AUX
    # EXX
    Exx__IndexOverlayed,
    Exx__IndexNotSet,
    Exx__ItemNotExists,
    Exx__StartOuterNONE_UsedInStackByRecreation,
)
from .breeder_1_str_2_stack import (
    # BASE
    BreederStrStack,
    # AUX
    BreederStrStack_Example,
    BreederStrStack_Example__BestUsage
    # TYPES
    # EXX
)
from .breeder_2_objects import (
    # BASE
    BreederObjectList,
    # AUX
    BreederObjectList_GroupType,
    # TYPES
    TYPE__BREED_RESULT__ITEM,
    TYPE__BREED_RESULT__GROUP,
    TYPE__BREED_RESULT__GROUPS,
    # EXX
    Exx__BreederObjectList_GroupsNotGenerated,
    Exx__BreederObjectList_GroupNotExists,
    Exx__BreederObjectList_ObjCantAccessIndex,
)
# ---------------------------------------------------------------------------------------------------------------------
from .singleton import (
    # BASE
    SingletonManagerBase,
    SingletonMetaCallClass,
    SingletonByCallMeta,
    SingletonByNew,
    # AUX
    # TYPES
    # EXX
    Exx_SingletonNestingLevels,
)


# =====================================================================================================================



"""
# TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
# TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
# TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
# TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
# TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
# TODO: NEED DEPRECATE IT!!! APPLY IN VALIDCHAINS!!!
but maybe leave it on!
"""

from typing import *

from base_aux.funcs import *
from base_aux.classes import *


# =====================================================================================================================
TYPE__RESULT_CUM_STEP = Union[bool, Any, "Valid"]     # DONT USE CALLABLES!!! Valid is most PREFIRED! specially in case of EXX!!
TYPE__RESULT_CUM_STEPS = Union[TYPE__RESULT_CUM_STEP, list[TYPE__RESULT_CUM_STEP]]


# =====================================================================================================================
class ResultStep:
    """
    # TODO: this is just as idea! but just create clear correct log msgs and this is will be enough!
    GOAL
    ----
    item for steps in sequence

    CREATED SPECIALLY FOR
    ---------------------
    for ResultCum as internal aux

    CONSTRAINTS
    -----------

    BEST USAGE
    ----------
    """
    pass


# =====================================================================================================================
class ResultCum:
    """
    GOAL
    ----
    apply cumulated result with log lines

    CREATED SPECIALLY FOR
    ---------------------
    testplans

    REPLACE: ResultExpected

    CONSTRAINTS
    -----------
    1. dont use callables in steps! only final result (as expression) or Valid as most preferred
    2. if no steps - return True!

    BEST USAGE
    ----------
        # some example
        def meth() -> ResultCum:
            result_cum = ResultCum()

            result_step = 1+1 == 2
            result_cum.result__apply_step(result_step)
            result_cum.log_lines__add('check 1+1 == 2')

            result_step = 1+2 == 4
            result_cum.result__apply_step(result_step)
            result_cum.log_lines__add('check 1+2 == 4')

            result_cum.finish()
            return result_cum

    """

    TITLE: str = ""
    # COMMENT: str = ""

    result: bool = True
    finished: bool | None = None    # as help to see if process is finished - maybe need deprecate!

    LOG_LINES: list[str]
    STEP_HISTORY: list[tuple[TYPE__RESULT_CUM_STEP, bool]]     # as history! result+step

    def __init__(self, title: str = None):        #, steps=None):  #dont use steps here
        if title:
            self.TITLE = title

        self.clear()

    def clear(self) -> None:
        self.result = True
        self.finished = None

        self.LOG_LINES = []
        self.STEP_HISTORY = []

    def result__apply_step(self, step: TYPE__RESULT_CUM_STEPS, cumulate: bool | Any = True, msg: Union[None, str, list[str]] = None) -> bool:
        """

        :param step:
        :param cumulate: False - if you dont mind step result! but want to keep it in log
        :param msg:
        :return:
        """
        if isinstance(step, list):
            # LIST ------------------------------------------
            for step_i in step:
                if not self.result__apply_step(step_i, cumulate):
                    break
        else:
            # SINGLE ------------------------------------------
            # TODO: if isinstance(step, ResultCum):   - what do i need to do here?? decide!
            if isinstance(step, Valid):
                if not step.finished:
                    step.run()
                self.log_lines__add(str(step))

        if msg:
            self.log_lines__add(msg)

        step_result = bool(step)
        self.STEP_HISTORY.append((step_result, step))

        if bool(cumulate):
            if self.result is None:
                self.result = step_result
            else:
                self.result &= step_result

        return self.result

    def log_lines__add(self, new: Union[str, list[str]]) -> None:
        if isinstance(new, list):
            self.LOG_LINES.extend(new)
        if isinstance(new, str):
            self.LOG_LINES.append(new)

    def finish(self) -> None:
        self.finished = True

    def __bool__(self) -> bool:
        return bool(self.result)

    def __str__(self):
        result = f"{self.__class__.__name__}(result={self.result},finished={self.finished},TITLE={self.TITLE},LOG_LINES:"
        for i, line in enumerate(self.LOG_LINES):
            result += f"\n{i:->5d}:{line}"
        return result

    def __repr__(self):
        return str(self)

    def step_last__get(self) -> Any | NoReturn:
        return self.STEP_HISTORY[-1][1]

    def log_last__get(self) -> str | None:
        try:
            return self.LOG_LINES[-1]
        except:
            pass


# =====================================================================================================================



from typing import *

from .valid_1_base import Valid
from base_aux.funcs import ValueNotExist


# =====================================================================================================================
class Exx__ValueNotInVariants(Exception):
    pass


class Exx__VariantsIncompatible(Exception):     # TODO: seems need to deprecate it! it is not important!
    pass


# =====================================================================================================================
TYPE__VARIANT = Union[str, Any]
TYPE__VARIANTS = Iterable[TYPE__VARIANT] | ValueNotExist


# =====================================================================================================================
class ValueVariants:
    """
    used to keep separated VALUE and measure unit

    GOAL
    ----
    1. get first associated value
    2. validate item by variants
    """
    # TODO: combine with ValueUnit - just add ACCEPTABLE(*VARIANTS) and rename UNIT just as SUFFIX!

    # SETTINGS -----------------------
    CASE_INSENSITIVE: bool = True
    VARIANTS: TYPE__VARIANTS = ValueNotExist
    VALUE_DEFAULT: Any = ValueNotExist

    # DATA ---------------------------
    __value: Any = ValueNotExist

    def __init__(self, value: Union[str, Any] = ValueNotExist, variants: TYPE__VARIANTS = ValueNotExist, case_insensitive: bool = None):
        """
        """
        if case_insensitive is not None:
            self.CASE_INSENSITIVE = case_insensitive

        self._variants_apply(variants)

        if value != ValueNotExist:
            self.VALUE = value
            self.VALUE_DEFAULT = self.VALUE

        self._variants_apply(variants)  # need secondary!!!

    def _variants_apply(self, variants: set[Union[str, Any]] | ValueNotExist = ValueNotExist) -> None:
        if variants is not ValueNotExist:
            self.VARIANTS = variants

        if self.VARIANTS is ValueNotExist:
            if self.VALUE is not ValueNotExist:
                self.VARIANTS = {self.VALUE, }
            # else:
            #     self.VARIANTS = set()

    def __str__(self) -> str:
        return f"{self.VALUE}"

    def __repr__(self) -> str:
        """
        used as help
        """
        return f"{self.VALUE}{self.VARIANTS}"

    def __eq__(self, other):
        if isinstance(other, ValueVariants):
            if other.VALUE == ValueNotExist:
                return self.VALUE in other
            else:
                other = other.VALUE

        if self.VALUE == ValueNotExist:
            return self.value_validate(other)

        # todo: decide is it correct using comparing by str()??? by now i think it is good enough! but maybe add it as parameter
        if self.CASE_INSENSITIVE:
            return (self.VALUE == other) or (str(self.VALUE).lower() == str(other).lower())
        else:
            return (self.VALUE == other) or (str(self.VALUE) == str(other))

    def __len__(self):
        return len(self.VARIANTS or [])

    def __iter__(self):
        yield from self.VARIANTS

    def __contains__(self, item) -> bool:
        """
        used to check compatibility
        """
        return self.value_validate(item)

    @property
    def VALUE(self) -> Any:
        return self.__value

    @VALUE.setter
    def VALUE(self, value: Any) -> Optional[NoReturn]:
        variant = self.value_get_variant(value)
        if variant != ValueNotExist:
            self.__value = variant
        else:
            raise Exx__ValueNotInVariants()

    def value_get_variant(self, value: Any) -> TYPE__VARIANT | ValueNotExist:
        for variant in self.VARIANTS:
            if Valid.compare_doublesided__bool(variant, value):
                return variant

            if self.CASE_INSENSITIVE:
                result = str(variant).lower() == str(value).lower()
            else:
                result = str(variant) == str(value)
            if result:
                return variant

        return ValueNotExist

    def value_validate(self, value: Any) -> Any | None:
        return self.value_get_variant(value) != ValueNotExist

    def reset(self) -> None:
        """
        set VALUE into default only if default is exists!
        """
        if self.VALUE_DEFAULT != ValueNotExist:
            self.VALUE = self.VALUE_DEFAULT


# =====================================================================================================================
